// Copyright 2006, Sun Microsystems, Inc.  All Rights Reserved.

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <poll.h>
#include "alloca.h"
#include <pthread.h>
#include <unistd.h>
#include "tl.h"
#include "rb.h"
#include "x86_sync_defns.h"

#define MAX_THREADS 128

static int Verbose = 1;
static volatile unsigned int threads_alive = 0;
static pthread_mutex_t startup_mutex[1];
static volatile int can_start = 0;
static volatile int stop_now = 0;
static KVMap *ht;

static int arg_inserts = 0;
static int arg_deletes = 0;
static int arg_updates = 0 ;
static int arg_initial_size = 10000;
static int uniq ;           // Starting population
static unsigned int arg_opgroup = 1;
static unsigned int arg_n_threads = 8;
static unsigned int arg_time_msec = 2000;
//static int arg_separate_readers = 0;
static char * ExecutableName = NULL ;
int arg_max_log_size = 1000;
int arg_max_handled_requests = 1000;
int arg_max_wait_for_slave = 999999;
int arg_no_slaves = 0;
static int arg_range = 1000000;

typedef struct {
    int pid;
    Thread * Self ;
    volatile int State ;
    volatile int nUpdates ;
    volatile int nDeletes ;
    volatile int nInserts ;
    volatile int nLookups ;
    volatile int nMisses ;
    double pad [64] ;
}
thread_data_t;

thread_data_t thread_data[MAX_THREADS];



#define get_time_u64() \
({ 	struct timeval x; \
	gettimeofday(&x, NULL); \
	((time_u64)(x.tv_sec * 1000000) + (time_u64)x.tv_usec); })



// TODO: replace low-quality Linear Congruential LCRNG
// with a Marsaglia-XOR or Park-Miller.
int SimpleLCRng (int * seed) {
    *seed = ((*seed) * 1103515245 + 12345)  ;
    return ((*seed) >> 16) & 0x7FFF ;
}

int ParkMillerRNG(int *seed0) {
    const int a =      16807;
    const int m = 2147483647;
    const int q =     127773;  /* m div a */
    const int r =       2836;  /* m mod a */
    int seed = *seed0;
    int hi   = seed / q;
    int lo   = seed % q;
    int test = a * lo - r * hi;
    if (test > 0)
        seed = test;
    else
        seed = test + m;
    *seed0 = seed;
    return seed;
}

//static volatile int gseed = 1 ;

#define TLRand(sa) MarsagliaXOR(sa)

void * Worker (void *arg) {
    int TallyMisses = 0 ;
    int TallyUpdates = 0;
    int TallyInserts = 0 ;
    int TallyDeletes = 0 ;
    int TallyLookups = 0 ;
    thread_data_t *pdata = (thread_data_t *)arg;
    Thread * Self ;
    int seed = (long)(int*)&seed+(int)RDTICK();
    seed |=1 ;
    pdata->Self = Self = TxNewThread () ;
    alloca ((pdata->pid * 7297) & 0xFFFF) ;
    pthread_mutex_lock(startup_mutex);
    if (pdata->pid == 0) {
        int i;
        for (i = 0; i < arg_initial_size; i++) {
            int key = TLRand(&seed) % arg_range ;
            if (!kv_contains (Self, ht, key)) {
                ;
                kv_put(Self, ht, key, key);
                ++uniq ;
            }
        }
        printf ("Initialized %d unique of %d\n", uniq, arg_initial_size) ;
    }
    ++threads_alive;
    pthread_mutex_unlock(startup_mutex);

    // TODO-FIXME: use a proper sense-reversing consensus barrier.
    while (!can_start) { }

    while (!stop_now) {
        int op = TLRand (&seed) % 100 ;
        int i;
        int opgroup = TLRand (&seed) % arg_opgroup ;
        if (arg_opgroup == 1)
            opgroup = 1;
        if (op < arg_inserts) {
            for (i = 0; i < opgroup; i++) {
                int key = TLRand(&seed) % arg_range ;
                kv_insert(Self, ht, key, key );
                TallyInserts ++ ;
            }
        } else if (op >= arg_inserts && op < (arg_inserts+arg_updates)) {
            for (i = 0; i < opgroup; i++) {
                int key = TLRand(&seed) % arg_range ;
                kv_put (Self, ht, key, TLRand(&seed)) ;
                TallyUpdates ++ ;
            }
        } else if (op >= 100 - arg_deletes) {
            for (i = 0; i < opgroup; i++) {
                int key = TLRand(&seed) % arg_range ;
                kv_delete(Self, ht, key);
                TallyDeletes ++ ;
            }
        } else {
            // Consider: the key distribution shouldn't be random.
            // Use some non-uniform random number
            // Idea: key = (key * key)/ arg_range
            for (i = 0; i < opgroup; i++) {
                int key = TLRand(&seed) % arg_range ;
                int hit = kv_get(Self, ht, key);
                TallyLookups ++ ;
                if (!hit)
                    TallyMisses ++ ;
            }
        }
    }

    pdata->nMisses  = TallyMisses ;
    pdata->nUpdates = TallyUpdates ;
    pdata->nDeletes = TallyDeletes ;
    pdata->nLookups = TallyLookups ;
    pdata->nInserts = TallyInserts ;
    return NULL;
}


#define RANGE 5000
static int map [RANGE] ;    // Tracks the ht - key:value map

int SelfTest (Thread * Self, KVMap * ht) {
    int i ;
    int seed = RDTICK() ;
    printf ("(1) Single-threaded self-test\n") ;
    for (i = 0 ; i < 2000000 ; i++ ) {
        int op = TLRand(&seed) % 100 ;
        if (op < 20) {
            int key = TLRand (&seed) % RANGE ;
            kv_delete (Self, ht, key) ;
            map [key] = 0 ;
            if (kv_get (Self, ht, key) != 0) {
                printf ("%d: delete read-back %d\n", key, kv_get (Self, ht, key)) ;
            }
        } else
            if (op >= 20 && op < 50) {
                int key = TLRand (&seed) % RANGE ;
                int val = TLRand (&seed) ;
                kv_put (Self, ht, key, val) ;
                map [key] = val ;
            } else {
                int key = TLRand (&seed) % RANGE ;
                if (kv_get(Self, ht, key) != map[key]) {
                    printf ("%d: map=%d ht=%d\n", key, map[key], kv_get(Self, ht, key)) ;
                }
            }
    }
    printf ("(2) read-back\n") ;
    for (i = 0 ; i < RANGE ; i++) {
        if (kv_get (Self, ht, i) != map[i]) {
            printf ("%d: map=%d ht=%d\n", i, map[i], kv_get(Self, ht, i)) ;
        }
    }
    printf ("(3) integrity check\n") ;
    kv_verify (ht, 1) ;
    printf ("(4) completed\n") ;
    return 0 ;
}

int do_test() {
    pthread_t* threads;
    pthread_attr_t thread_attr [1];
    unsigned i;
    time_u64 start_time, end_time;

    threads = (pthread_t *)malloc(sizeof(pthread_t) * arg_n_threads);
    memset (threads, 0, sizeof(*threads) * arg_n_threads) ;
    pthread_attr_init(thread_attr);
    if (1) {
        pthread_attr_setschedpolicy(thread_attr, SCHED_RR);
        pthread_attr_setscope(thread_attr, PTHREAD_SCOPE_SYSTEM) ;
    }
    pthread_mutex_init(startup_mutex, NULL);

    // Run a quick smoke-test
    Thread * Self = TxNewThread () ;
    if (getenv ("SELFTEST") != NULL) {
        SelfTest (Self, ht) ;
    }

    if (Verbose) {
        printf ("  ") ;
        printf ("V%d ",             kv_verify (ht, 0)) ;
        printf ("INSERT(55)=%d "  , kv_insert (Self, ht, 55, 55)) ;
        printf ("INSERT(55)=%d "  , kv_insert (Self, ht, 55, 55)) ;
        printf ("GET(55)=%d "     , kv_get    (Self, ht, 55)) ;
        printf ("CONTAINS(55)=%d ", kv_contains (Self, ht, 55)) ;
        printf ("DELETE(55)=%d "  , kv_delete (Self, ht, 55)) ;
        printf ("DELETE(55)=%d "  , kv_delete (Self, ht, 55)) ;
        printf ("GET(55)=%d "     , kv_get    (Self, ht, 55)) ;
        printf ("CONTAINS(55)=%d ", kv_contains (Self, ht, 55)) ;
        printf ("V%d\n",            kv_verify (ht, 0)) ;
        printf ("  ") ;
        printf ("Insert ") ;
        int i ;
        for (i = 0 ; i < 20 ; i++)
            kv_insert (Self, ht, i ^ 0x5A, i ^ 0x5A) ;
        int fail = 0 ;
        for (i = 0 ; i < 20 ; i++) {
            kv_delete (Self, ht, i ^ 0x5A) ;
            if (fail == 0 && !kv_verify (ht, 0)) {
                ++fail ;
                printf ("fail >> %d (%d)\n", i, (i ^ 0x5A)) ;
            }
        }
        if (kv_verify (ht, 0) <= 0)
            printf ("VERIFY FAILURE!\n") ;

        printf (" " ) ;
        printf ("GET(1000)=%d "   , kv_get(Self, ht, 1000)) ;
        printf ("SET(1000,1)=%d " , kv_put(Self, ht, 1000, 1)) ;
        printf ("GET(1000)=%d "   , kv_get(Self, ht, 1000)) ;
        printf ("SET(1000,2)=%d " , kv_put(Self, ht, 1000, 2)) ;
        printf ("GET(1000)=%d "   , kv_get(Self, ht, 1000)) ;
        printf ("V%d\n", kv_verify (ht, 0)) ;
        printf ("\n") ;
    }


    for (i = 0; i < arg_n_threads; i++) {
        int res;
        thread_data[i].pid = i;
        res = pthread_create(&threads[i], thread_attr, Worker , (void *)(thread_data + i));
        if (res != 0) {
            printf("error creating thread, error: %s\n", strerror(res));
            return -1;
        }
    }
    // Wait for all the threads to be running.
    while (threads_alive < arg_n_threads) {
        poll (NULL,0,1);
    }

    fprintf(stderr, "starting...\n");
    start_time = get_time_u64();
    can_start = 1;
    usleep(arg_time_msec * 1000);
    stop_now  = 1;
    end_time = get_time_u64();
    printf ("shutdown...") ;

    /* Wait for all the threads to shut down.
    */
    for (i = 0; i < arg_n_threads; i++) {
        pthread_join(threads[i], NULL);
    }
    free (threads);

    // TODO: report stddev, median, average for (I,D,L,U)

    printf("results:\n");
    int all_inserts = 0 ;
    int all_misses  = 0 ;
    int all_updates = 0 ;
    int all_deletes = 0 ;
    int all_lookups = 0 ;
    int all_aborts  = 0 ;
    int nOps = 0 ;
    int64_t lds, sts ;
    lds = sts = 0 ;
    for (i = 0; i < arg_n_threads; i++) {
        printf("(%d, %d, %d, %d) ",
               thread_data[i].nInserts,
               thread_data[i].nDeletes,
               thread_data[i].nUpdates,
               thread_data[i].nLookups) ;
//        lds         += thread_data[i].Self->TxLD ;
//        sts         += thread_data[i].Self->TxST ;
//        all_aborts  += thread_data[i].Self->Aborts ;
        all_misses  += thread_data[i].nMisses ;

        all_inserts += thread_data[i].nInserts ;
        all_updates += thread_data[i].nUpdates ;
        all_lookups += thread_data[i].nLookups ;
        all_deletes += thread_data[i].nDeletes ;

        if ((i % 3) == 2)
            printf ("\n") ;
    }
    printf ("\n") ;
    nOps = all_inserts + all_updates + all_deletes + all_lookups ;

    // Crude - iterate over the key space to determine the population (cardinality)
    // of the RB-tree.  A better way would be to augment the recursive _verify()
    // routine to return the population.
    int upop = 0 ;
    for (i = 0 ; i < arg_range ; i++) {
        upop += kv_contains (Self, ht, i) != 0 ;
    }

    printf ("TEST: %s %dT %d msecs ins=%%%d del=%%%d ups=%%%d isize=%d (initpop=%d) range=%d\n",
            TxDescribe(),
            arg_n_threads, arg_time_msec, arg_inserts, arg_deletes, arg_updates, arg_initial_size,
            uniq, arg_range) ;

    // uniq is the initial population.
    // TODO: We should report the final population, too.

    printf ("RESULTS: Dur=%d pop=%d U=%d I=%d D=%d L=%d (Misses=%d) TOTAL=%d\n",
            (int)((end_time - start_time) / 1000),
            upop,
            all_updates, all_inserts, all_deletes, all_lookups, all_misses, nOps) ;
    printf ("RESULTS: TxLDs=%ld TxSTs=%ld\n", (long)lds, (long)sts) ;
    printf ("SUMMARY: %s %s %dT I%d D%d U%d L%d (%d,%d) ABO=%d pop=%d -> %d Ops\n",
            ExecutableName,
            TxDescribe(),
            arg_n_threads, arg_inserts, arg_deletes, arg_updates,
            100 - (arg_inserts + arg_deletes + arg_updates),
            arg_initial_size, arg_range, all_aborts, upop, nOps) ;

    // Run a post-mortem verification pass.
    // There are no concurrent operations on the ht.
    // TODO: implement a non-transactional verify we can run when
    // the ht is stable and queisced.
    // Verify post-conditions.
    printf ("VERIFY=%d\n", kv_verify (ht, 1)) ;
    if (0) {
        printf ("INSERT(55)=%d "   , kv_insert (Self, ht, 55, 55)) ;
        printf ("INSERT(55)=%d "   , kv_insert (Self, ht, 55, 55)) ;
        printf ("CONTAINS(55)=%d " , kv_contains (Self, ht, 55)) ;
        printf ("DELETE(55)=%d "   , kv_delete (Self, ht, 55)) ;
        printf ("DELETE(55)=%d "   , kv_delete (Self, ht, 55)) ;
        printf ("CONTAINS(55)=%d " , kv_contains (Self, ht, 55)) ;
        printf ("VERIFY=%d\n"      , kv_verify(ht, 0)) ;
    }
    TxShutdownStats() ;

    // consider: enable MSA?
    struct rusage ru ;
    getrusage (RUSAGE_SELF, &ru) ;
    printf ("User=%ld System=%ld msecs\n",
            ((ru.ru_utime.tv_sec * 1000000L) + ru.ru_utime.tv_usec)/1000L,
            ((ru.ru_stime.tv_sec * 1000000L) + ru.ru_stime.tv_usec)/1000L) ;

return 1;
}


int process_arguments(int argc, char *argv[]) {
    --argc ;
    ++argv ;
    while (argc > 0 && **argv == '-') {
        char * p = *(argv++) + 1 ;
        --argc ;
        switch (*(p++)) {
        case 't':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_time_msec = strtol (p, NULL, 0) ;
            break ;
        case 's':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_initial_size = strtol (p, NULL, 0) ;
            break ;
        case 'n':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_n_threads = strtol (p, NULL, 0) ;
            break ;
        case 'u':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_updates = strtol (p, NULL, 0) ;
            break ;
        case 'i':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_inserts = strtol (p, NULL, 0) ;
            break ;
        case 'd':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_deletes = strtol (p, NULL, 0) ;
            break ;
        case 'g':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_opgroup = strtol (p, NULL, 0) ;
            break ;
        case 'r':
            if (*p == 0 && argc > 0) {
                p = *(argv++) ;
                --argc ;
            }
            arg_range = strtol (p, NULL, 0) ;
            break ;
        default:
            printf ("UNKNOWN SWITCH: %s\n", p-1) ;
            return 0;
        }
    }
    return 1;
}

int main(int argc, char *argv[]) {
    time_t seed;
    kv_init();
    seed = 1001;

    setbuf (stdout, NULL) ;
    ExecutableName = argv[0] ;
    //TxOnce () ;

    if (!process_arguments(argc, argv)) {
        printf("Usage:\ntest -t time_msec "
               "-n nthreads -i inserts -d deletes -s initial_size "
               "-r range\n");
        return -1;
    }

    srand(seed);
    ht = kv_create(-1, NULL);
    do_test();
    return 0;
}

