
#include "trace_log.h"

#ifdef ENABLE_TRACE
#warning TRACING is enabled

#define TRACE_MAX_THREADS 100

#define MAX_EVENTS 4000



typedef struct _TraceLogEvent {
    long event_ts;
    int thread_id;
    int step;
    TraceEventName type;
    intptr_t arg1, arg2;
    intptr_t const volatile *addr;

}
TraceLogEvent;


static volatile TraceLogEvent GlobalTraceLog[MAX_EVENTS];
static volatile int lock_var=0;
static volatile long global_event_clock=0;

char * TraceEventString[] = {
                                (char*)"_tl_txstart    ",
                                (char*)"_tl_txabort    ",
                                (char*)"_tl_txcommit   ",
                                (char*)"_tl_sigsegf    ",
                                (char*)"_tl_rscoherent ",
                                (char*)"_tl_txload     ",
                                (char*)"_tl_txvfy      ",
                                (char*)"_tl_txstore    ",
                                (char*)"_tl_txsterilize",
                                (char*)"_tl_LK_ldlock  ",
                                (char*)"_tl_LK_stlock  ",
                                (char*)"_tl_LK_caslock ",
                                (char*)"_tl_LK_ldaddr  ",
                                (char*)"_tl_LK_staddr  ",
                                (char*)"_tl_KV_get     ",
                                (char*)"_tl_KV_put     ",
                                (char*)"_tl_KV_delete  " };

// -----------------------------------------------------------
static inline void spin_lock(int volatile * const lock) {
    for(;;) {
        while(*lock==1) {
            sched_yield();
        }
        if(CASIO(lock, 0,1)==0) {
            break;
        }
    }
    assert(*lock==1);

}

static inline int spin_trylock(int volatile * const lock) {
    if(*lock==0 && CASIO(lock, 0, 1)==0) {
        assert(*lock==1);
        return 1;
    }

    return 0;
}


static inline void spin_unlock(int volatile * const lock) {
    assert(*lock==1);
    *lock=0;
}

static inline void CAS_SPIN_SET(uintptr_t * const addr, uintptr_t const val) {
    uintptr_t old_val;
    do {
        old_val=*addr;
    } while(old_val!=CASPO(addr, old_val, val));
}

static inline long IncClock() {
    long v;
    for (v = global_event_clock; CASPO(&global_event_clock, v, v+1) != v; v = global_event_clock)
        ;

    assert(v+1!=0);
    return v;
}

// --------------------------------------------

// -----------------------------------------------------------


inline void TraceEvent(int thread_id, int step, enum _TraceEventName type, intptr_t const volatile *addr, intptr_t arg1, intptr_t arg2) {
    long my_clock = IncClock();
    long pos=my_clock%MAX_EVENTS;

    GlobalTraceLog[pos].event_ts=my_clock;
    GlobalTraceLog[pos].thread_id=thread_id;
    GlobalTraceLog[pos].step=step;
    GlobalTraceLog[pos].type=type;
    GlobalTraceLog[pos].addr=addr;
    GlobalTraceLog[pos].arg1=arg1;
    GlobalTraceLog[pos].arg2=arg2;
}

inline void TraceTxStart(int thread_id, int step, vwLock rv) {
    TraceEvent(thread_id, step, _tl_txstart, NULL, (intptr_t)0, (intptr_t)rv);
}

inline void TraceTxCommit(int thread_id, int step) {
    TraceEvent(thread_id, step, _tl_txcommit, NULL, (intptr_t)0, (intptr_t)0);
}

inline void TraceTxAbort(int thread_id, int step) {
    TraceEvent(thread_id, step, _tl_txabort, NULL, (intptr_t)0, (intptr_t)0);
}

inline void TraceTxLoad(int thread_id, int step, intptr_t const volatile * addr) {
    TraceEvent(thread_id, step, _tl_txload, addr, (intptr_t)0, (intptr_t)0);
}

inline void TraceTxStore(int thread_id, int step, intptr_t const volatile * addr, intptr_t val) {
    TraceEvent(thread_id, step, _tl_txstore, addr, (intptr_t)0, val);
}

inline void TraceTxSterilize(int thread_id, int step, intptr_t const volatile * addr) {
    TraceEvent(thread_id, step, _tl_txsterilize, addr, (intptr_t)0, (intptr_t)0);
}

inline void TraceSIGSEGF(int thread_id, int step, intptr_t const volatile * addr) {
    TraceEvent(thread_id, step, _tl_sigsegf, addr, (intptr_t)0, (intptr_t)0);
}

inline void TraceRSCoherent(int thread_id, int step, intptr_t const volatile * addr, intptr_t val) {
    TraceEvent(thread_id, step, _tl_rscoherent, addr, (intptr_t)0, val);
}

static inline void __PrintTrace(FILE * tracelog) {


    assert(global_event_clock!=0);
    int i=0;
    int my_clock, end_clock;
    end_clock = my_clock = global_event_clock-1;
    do {
        TraceLogEvent volatile * ptr = &GlobalTraceLog[my_clock%MAX_EVENTS];
        if(ptr->event_ts>=0) {
            fprintf(tracelog, " %5d -- id: %2d aft: %3d type: %16s addr: %10lx arg1: %10lx arg2: %10lx\n",
                    i, ptr->thread_id, ptr->step, TraceEventString[ptr->type], (long)ptr->addr, (long)ptr->arg1, (long)ptr->arg2);
        }

        my_clock--;
        i++;
    } while(my_clock>=0 && my_clock%MAX_EVENTS != end_clock%MAX_EVENTS);
}


void PrintTrace() {
    char tracelog_file_name[200];
    spin_lock(&lock_var);

    sprintf(tracelog_file_name, "global_trace.log");
    FILE * tracelog = fopen(tracelog_file_name, "w");
    fprintf(tracelog, "GLOBAL TRACELOG:\n");
    __PrintTrace(tracelog);
    fclose(tracelog);

    spin_unlock(&lock_var);

}

void TraceInit() {
    int i;
    for(i=0;i<MAX_EVENTS;i++) {
        GlobalTraceLog[i].event_ts=-1;
    }
}

#endif



